#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  litho3d_stl.py
#  
#  Copyright 2015 Philippe Bonnet <phil.bonnet@laposte.net>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  

import getopt, sys
from PIL import Image, ImageFilter


class STLConverter():
    
    def __init__(self):
        self.output = None
        self.crop = None
        self.precision = None
        self.preserve = False
        self.maxlen = None
        self.coef = 1.0
        self.number = 8
        self.thickness = 0.15
        self.show = False

    def set_options(self, opts):
        for o, a in opts:
            if o in ("-h", "--help"):
                self.usage()
                sys.exit()
            elif o in ("-o", "--output"):
                self.output = a
            elif o in ("-c", "--crop"):
                self.crop = [int(i) for i in a.split(",")]
            elif o in ("-P", "--precision"):
                self.precision = float(a)
                self.preserve = True
            elif o in ("-p", "--preserve"):
                self.preserve = True
            elif o in ("-l", "--len"):
                self.maxlen = float(a)
            elif o in ("-n", "--number"):
                self.number = int(a)
            elif o in ("-t", "--thickness"):
                self.thickness = float(a)
            elif o in ("-s", "--showonly"):
                self.show = True
            else:
                assert False, "unhandled option"

    def load_image(self, filename):
        try:
            self.im = Image.open(filename)
        except:
            return 
        
        if self.crop:
            self.im = self.im.crop((self.crop[0],
                                    self.crop[1],
                                    self.crop[0]+self.crop[2],
                                    self.crop[1]+self.crop[3]))
        w,h = self.im.size
        if self.precision:
            self.im = self.im.resize((int(w*self.precision),int(h*self.precision)))
        if self.maxlen:
            if not self.preserve:
                self.coef = max(w,h) / self.maxlen / (150/25.4)    # Def max dpi = 150
                self.im = self.im.resize((int(w/self.coef),int(h/self.coef)))
                w,h = self.im.size
            self.coef = self.maxlen / max(w,h)
        self.im = self.im.convert("L")
        
    def create_stl(self, filename='qrcode.'):
        if not(self.output):
            #self.output = ".".join(filename.split("/")[-1:][0].split(".")[:-1])+"_auto.stl"
            self.output = ".".join(filename.split(".")[:-1])+"_auto.stl"

        self.im = self.im.point(lambda i: (self.number-int(i/256.0*self.number) + 0))
        w,h = self.im.size
        fic = open(self.output,"w")
        self.output = None

        fic.write(
"""solid litho_stl
"""
        )
        
        # Vertices
        vertices = []
        for i in range(w):
            for j in range(h):
                vertices += [[i*self.coef, (h-j-1)*self.coef, self.im.getpixel((i, j))*self.thickness]]
        """
            Ajouter les points pour construire le socle
            à la verticale des bords de l'image
        """
        for j in range(h-1):
            i=0
            vertices += [[i*self.coef, (h-j-1)*self.coef, -self.thickness]]
        for i in range(w-1):
            j=h-1
            vertices += [[i*self.coef, (h-j-1)*self.coef, -self.thickness]]
        for j in range(h-1,0,-1):
            i=w-1
            vertices += [[i*self.coef, (h-j-1)*self.coef, -self.thickness]]
        for i in range(w-1,0,-1):
            j=0
            vertices += [[i*self.coef, (h-j-1)*self.coef, -self.thickness]]
        
        # Faces
        for i in range(1,w):
            for j in range(1,h):
                n=(i*h)+j
                fic.write(self.add_facet([vertices[n-h-1], vertices[n-h-0], vertices[n-0-1]]))
                fic.write(self.add_facet([vertices[n-h-0], vertices[n-0-0], vertices[n-0-1]]))
        
        """
            Constituer la liste des points en bordure de l'image
            puis la liste des points ajoutés à leur verticale pour le socle
        """
        top_lst  = [i for i in range(h-1)]
        top_lst += [h*i-1 for i in range(1,w)]
        top_lst += [h*w-1-i for i in range(h-1)]
        top_lst += [h*i for i in range(w-1,0,-1)]
        
        d=h*w   # Delta debut des points du socle
        bot_lst  = [d+i for i in range(h-1)]
        bot_lst += [d+h-1+i for i in range(w-1)]
        bot_lst += [d+h+w-2+i for i in range(h-1)]
        bot_lst += [d+h+w+h-3+i for i in range(w-1)]

        lst = zip(top_lst, bot_lst)
        prv = lst[0]
        lst.remove(lst[0])
        lst.append(prv)
        
        for t,b in lst:
            tp, bp = prv
            fic.write(self.add_facet([vertices[b], vertices[tp], vertices[bp]]))
            fic.write(self.add_facet([vertices[b], vertices[t],  vertices[tp]]))
            prv = (t, b)
            
        B = bot_lst[0]
        bot_lst.remove(bot_lst[0])
        C = bot_lst.pop()
        while bot_lst:
            A = bot_lst[0]
            bot_lst.remove(bot_lst[0])
            fic.write(self.add_facet([vertices[A], vertices[B], vertices[C]]))
            B = A
            A = bot_lst.pop()
            fic.write(self.add_facet([vertices[A], vertices[B], vertices[C]]))
            C = A
        fic.write(
"""endsolid litho_stl
"""
        )

    def add_facet(self, vert):
        v1, v2, v3 = vert
        v1x, v1y, v1z = v1
        v2x, v2y, v2z = v2
        v3x, v3y, v3z = v3
        res = "facet normal 0.0 0.0 0.0\n"
        res += "   outer loop\n"
        res += "     vertex %e %e %e\n" %(v1x, v1y, v1z)
        res += "     vertex %e %e %e\n" %(v2x, v2y, v2z)
        res += "     vertex %e %e %e\n" %(v3x, v3y, v3z)
        res += "   endloop\n"
        res += "endfacet\n"
        return res

    def usage(self):
        print """
    Usage : litho3d_stl.py [options] imagename.jpg [imagename2.jpg ...]
        
        options :
            [-h|--help] 
            [-o|--out filename.ply] -> ply output filename
            [-c|--crop] x,y,w,h     -> x,y upper left corner w,h sizes
            [-P|--precision]        -> more precision
            [-p|--preserve]         -> don't resize (60 dpi)
            [-l|--len]              -> max size in mm
            [-n|--number]           -> number of slices (8)
            [-t|--thickness] height -> thickness of each slices (0.15)
            [-s|--showonly
        """

def main():
    stlc = STLConverter()
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], 
                                    "ho:c:P:pl:n:t:s", 
                                    ["help", "output=", "crop=", 
                                     "precision", "preserve", "len=", 
                                     "number=", "thickness=", "showonly"])
    except getopt.GetoptError as err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        stlc.usage()
        sys.exit(2)
    
    stlc.set_options(opts)
    
    for fic in args:
        stlc.load_image(fic)
        if stlc.show:
            stlc.im.show()
        else:
            stlc.create_stl(fic)
    return 0

if __name__ == '__main__':
    main()
